/* client.c
 * MG server
 *
 */

#include "client.h"
#include "decor.h"
#include "commands.h"
#include "menu.h"
#include "network.h"
#include "request.h"
#include "schedule.h"
#include "windows.h"
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

int clients = 0;
client *_client_list = NULL;

client *_desktop = NULL;

client *desktop()
{
	return _desktop;
}

int make_desktop( client *c )
{
	client *t;
	if( desktop() )
		return -1;
	_desktop = c;
	c->window.x = (screen_width() - 
			     c->visual.mode.visible.x) / 2;
	c->window.y = (screen_height() - 
		       c->visual.mode.visible.y) / 2;
	c->border = c->window;
	c->border.next = NULL;
	unlink_client( c );

	t = _client_list;
	if( !t )
		_client_list = c;
	else
		while( t != NULL ){
			if( !t->next ){
				t->next = c;
				break;
			}
		}

	redraw_screen();
	printf("desktop bound\n");
	return 0;
}

client *client_list()
{
  return _client_list;
}

void init_mode(ggi_mode *mode)
{
  mode->frames = 0;
  mode->visible.x = 0;
  mode->visible.y = 0;
  mode->virt.x = 0;
  mode->virt.y = 0;
  mode->size.x = 0;
  mode->size.y = 0;
  mode->graphtype = GGI_AUTO;
}

void init_fields(client *c, int s)
{
	int i;
	
	c->flags = 0;
	c->parent = 0;
	c->private = 0;
	strncpy(c->name, "GGI App", 32);
	c->visual.target = target_visual();
	c->visual.pixelformat = *ggiGetPixelFormat(c->visual.target);
	c->visual.flags = 0;
	init_mode(&(c->visual.mode));
	c->event.mask = emAll;
	c->event.last = 0;
	c->event.poll_mask = 0;
	c->event.read_mask = 0;
	
	c->border.x = c->border.y = c->border.w = c->border.h = 0;
	c->window.x = c->window.y = c->window.w = c->window.h = 0;
	c->new = c->border;

	c->window_clip = NULL;
	c->border_clip = NULL;
	c->comm.packet.size = 0;
	c->comm.packet.type = 0;
	c->comm.out.size = 0;
	c->comm.out.type = 0;
	c->comm.sent = 0;
	c->comm.read = c->comm.sent = 0;
	c->comm.socket = s;
	for( i = 0; i < EVENTS; i++ )
		c->event.buffer[i].any.type = 0;
}



// return true if c should not be before t
int wrong_position( int place, client *c, client *t )
{
	if( t == _desktop )
		return 0;
	if( place == FIRST ){
		if( c->flags & FLOAT )
			return 0;
		if( t->flags & FLOAT )
			return 1;
		return 0;
	} else {
		if( t->flags & FLOAT )
			return 1;
		if( c->flags & FLOAT )
			return 0;
		return 1;
	}
}

void insert( client *c, int place )
{
	client *t, *n;
	n = t = client_list();
	while( t && wrong_position( place, c, t)){
		n = t;
		t = t->next;
	}
	if( n == t ){
		c->next = _client_list;
		_client_list = c;
	} else {
		c->next = n->next;
		n->next = c;
	}
}


// detach client from list
int raise_client( client *c )
{
	client *n = c->next;
	unlink_client( c );
	insert( c, FIRST );
	return (n != c->next);
}

int lower_client( client *c )
{
	client *n = c->next;
	unlink_client( c );
	insert( c, LAST );
	return (n != c->next);
}

void float_client( client *c )
{
	if( !(c->flags & FLOAT ) ){
		unlink_client( c );
		c->flags |= FLOAT;
		insert( c, FIRST );
	}
}

void unfloat_client( client *c )
{
	if( c->flags & FLOAT ){
		unlink_client( c );
		c->flags &= ~FLOAT;
		insert( c, FIRST );
	}
}

// {todo} add routines for moving in client_list;
void new_client(int s)
{
	MG_packet packet;
	client *c = malloc( sizeof( client) );
	while( s == 0 ){        // fd 0 is for menus
		int a = dup( s );
		close( s );
		s = a;
	}		
	init_fields(c,s);
	add_readable( c );
	init_packet(  &packet, Information );
	store_sint32( &packet, s );
	send_packet( c->comm.socket, &packet );

	insert( c, FIRST );
	// New client has focus
	set_focus( c );
	clients++;
	printf("new client at socket %d\n",s);
}

const ggi_pixelformat *mgPixelFormat( client *c )
{
	return &c->visual.pixelformat;
}

void destroy_client_data( client *c )
{
	if( !menu( c ) ){
		if( close( c->comm.socket ) )
			perror(" socket already closed\n");
	}
	clear_write( c );
	clear_readable( c );
	destroy_regions( &c->window_clip );
	destroy_regions( &c->border_clip );
	update_regions( &c->border );
//	printf(" redraw from destroy_client_data\n");
//	print_regions( &c->border );
	redraw_screen_area( &c->border );
	if( desktop() == c )
		_desktop = NULL;
	free( c );
	printf("client data is cleaned up\n");
}

void unlink_client(client *c)
{
	client *t = _client_list;
	if( c == t ){
		_client_list = c->next;
		c->next = NULL;
		return;
	}
	while( t != NULL ){
		if( t->next == c ){
			t->next = c->next;
			c->next = NULL;
			return;
		}
		t = t->next;
	}
}

// clean data structures, close sockets
void remove_client(client *c)
{
	unlink_client( c );
	if( c == get_focus() ){
		set_focus(NULL);
		update_focus();
	}
	destroy_client_data( c );
	clients--;
}

void add_client( int s )
{
    	int client_fd = accept_connection( s );
	if( client_fd == -1 ){
		return;
	}
	new_client(client_fd);
}

    
void process_clients( fd_set *readset, fd_set *writeset )
{
	client *t;
	client *c = client_list();
	
	while( c != NULL ){
		int kill = 0;
		t = c->next;
		if( FD_ISSET(c->comm.socket, writeset ) ){
			if( flush_buffer( &c->comm ) < 0 )
				kill = 1;
		}
		if( (c->comm.out.type == 0) &&
		    FD_ISSET(c->comm.socket, readset ) ){
			int err = load_packet( &c->comm );
			switch( err ){
			case 0:	kill = execute_requests( c );
				break;
			case 1: kill = 0;
				break;
			default: kill = 1;
				break;
			}
		}
		if( kill )
			remove_client( c );	   
		c = t;
	}
}





