/* client h
 * client management
 */
#ifndef CLIENT_H
#define CLIENT_H

#include "defines.h"
#include "regions.h"
#include "visual.h"

#define window_clip window.next
#define border_clip border.next

/* window flags:
   desktop        - window is stuck on the desktop /i.e in the bottom
   float          - window is floting on top of other windows
   stick          - window has a fixed position and cannot be moved
   tilt           - window has label and tag on the side 
   nodecor        - window has no decorations
   icon           - window is iconified
   invisible      - window is invisible
   responsive     - window accepts input without having focus
*/

#define DESKTOP       ( 1 << 0 )
#define FLOAT         ( 1 << 1 )
#define STICK         ( 1 << 2 )
#define TILT          ( 1 << 3 )
#define NODECOR       ( 1 << 4 )
#define INVISIBLE     ( 1 << 5 )
#define RESPONSIVE    ( 1 << 6 )

#define FIRST 0
#define LAST 1

typedef struct _client{
	struct _client *next;
	struct _client *parent;
	void *private;
	char name[32];
	communication comm;
	eventinfo event;
	visual_info visual;
	region window;
	region border;
	region new;
	unsigned int flags;
} client;



client *client_list();
void init_fields(client *c, int s);
void remove_client(client *c);
const ggi_pixelformat *mgPixelFormat( client *c );
void process_clients( fd_set *r, fd_set *w);
void add_client( int s );
client *desktop();
int make_desktop( client *c );
void unlink_client(client *c);
int raise_client( client *c );
int lower_client( client *c );
void float_client( client *c );
void unfloat_client( client *c );
void insert( client *c, int place );


#endif

