/* decor.c
 * MG server
 *
 * window decorations
 */

#include "decor.h"
#include "windows.h"

int border( client *c, int b )
{
	if( c->flags & NODECOR ){
		return 0;
	}
	if( c->flags & TILT ){
		switch( b ){
		case left: return 12;
		case right: return 2;
		case top: return 2;
		case bottom: return 2;
		}
	} else {
		switch( b ){
		case left: 
		case right: return 2;
		case top: return 12;
		case bottom: return 2;
		}
	}
	return 0;
}

ggi_pixel colormap[16];


ggi_pixel Map(ggi_visual_t vis, int rgb)
{
	ggi_color col;
	col.r = rgb;
	col.g = rgb;
	col.b = rgb;
	col.a = 0xffff;
	return ggiMapColor(vis, &col);
}
 
void init_palette(ggi_visual_t vis)
{
	colormap[0]  = Map(vis, 0xffff);
	colormap[1]  = Map(vis, 0xe000);
	colormap[2]  = Map(vis, 0xd000);
	colormap[3]  = Map(vis, 0xc000);
	colormap[4]  = Map(vis, 0xb000);
	colormap[5]  = Map(vis, 0xa000);
	colormap[6]  = Map(vis, 0x9000);
	colormap[7]  = Map(vis, 0x8000);
	colormap[8]  = Map(vis, 0x7000);
	colormap[9]  = Map(vis, 0x6000);
	colormap[10] = Map(vis, 0x5000);
	colormap[11] = Map(vis, 0x4000);
	colormap[12] = Map(vis, 0x3000);
	colormap[13] = Map(vis, 0x2000);
	colormap[14] = Map(vis, 0x1000);
	colormap[15] = Map(vis, 0x0000);
}


	
void Hline(client *c, int xoff, int  yoff, int w, ggi_pixel p)
{
	ggiSetGCForeground(c->visual.target, p);
	ggiDrawHLine(c->visual.target, c->border.x+xoff, c->border.y+yoff, w);
}

void Vline(client *c, int xoff, int  yoff, int h, ggi_pixel p)
{
	ggiSetGCForeground(c->visual.target, p);
	ggiDrawVLine(c->visual.target, c->border.x+xoff, c->border.y+yoff, h);
}

void Dot(client *c, int xoff, int yoff, ggi_pixel p)
{
	ggiSetGCForeground(c->visual.target, p);
	ggiPutPixel(c->visual.target, c->border.x+xoff, c->border.y+yoff, p);
}

void Box(client *c, int xoff, int yoff, int w, int h, ggi_pixel p)
{
	ggiSetGCForeground(c->visual.target, p);
	ggiDrawBox(c->visual.target, c->border.x+xoff, c->border.y+yoff, w,h);
}

void Text(client *c, int xoff, int yoff, const char *str, 
	  ggi_pixel f, ggi_pixel b)
{
	ggiSetGCBackground(c->visual.target, b);
	ggiSetGCForeground(c->visual.target, f);
	ggiPuts(c->visual.target, c->border.x+xoff, c->border.y+yoff, str);
}

void set_gc( client *c, int fg, int bg)
{
	c->visual.fg_color = colormap[fg];
	c->visual.bg_color = colormap[bg];
	ggiSetGCForeground( c->visual.target, colormap[fg] );
	ggiSetGCBackground( c->visual.target, colormap[bg] );
}



void new_draw_decorations( client *c )
{
	int base = 0;
	if( c->flags & NODECOR )
		return;
	if( c == get_focus() )
		base = 1;
	// top
	Hline( c, 0, 0, c->border.w-1, colormap[1]);
	Dot( c, c->border.w-1, 0, colormap[9]);
	Dot( c, border(c, left)-1, border(c, top)-1, colormap[8]);
	Hline( c, border(c, left), border(c, top)-1, c->window.w, colormap[9]);
	// bottom
	Dot( c, 0, c->border.h-1, colormap[9]);
	Hline( c, 1, c->border.h-1, c->border.w-1, colormap[12]);
	Dot( c, border(c, left)-1, c->border.h-2, colormap[4]);
	Hline( c, border(c, left), c->border.h-2, c->window.w+1, colormap[3]);
	// left
	Vline( c, 0, 1, c->border.h-2, colormap[1]);
	Vline( c, border(c, left)-1, border(c, top), c->window.h, colormap[7]);
	// right
	Vline( c, c->border.w-1, 1, c->border.h-2, colormap[12]);
	Vline( c, c->border.w-2, border(c, top), c->window.h, colormap[3]);

	if( c->flags & TILT ){
	} else {
		Box( c, 1, 1, c->border.w-2, border(c, top)-2, colormap[3]);
	}
	       
}

void draw_decorations( client *c )
{
	int i, x;
//	int x;
	int marker = 3;
	if( c->flags & NODECOR )
		return;
	if( c == get_focus() )
		marker = 5;
//	Box(c, 0,0, c->border.w, c->border.h, colormap[3]);
	Hline(c, 0, 0, c->border.w-1, colormap[1]);
	Dot(c, c->border.w-1, 0, colormap[9]);
	Box(c, 1, 1, c->border.w-2, 10, colormap[marker]);
//	x = MAX(c->border.w/2 - strlen(c->name)*4, 2);
	Text(c, 18, 2, c->name, colormap[15], colormap[marker]);
	for(i = 0; i < 4; i++ ){
		x = c->border.w - (8 - i);
		Dot(c, x, i+2, colormap[8] );
		Dot(c, x+1, i+2, colormap[2] );
	}
	Dot( c, x+1, 2, colormap[6]);
	for( i = 0; i < 8; i++ ){
		x = c->border.w - (12 - i );
		Dot(c, x, i+2, colormap[8]);
		Dot(c, x+1, i+2, colormap[2] );
	}


//	Box( c, 2, 2, 10, 8, colormap[10] );

	Vline(c, 14, 0, 12, colormap[8]);
	Vline(c, 15, 0, 12, colormap[2]);

//	Vline(c, 2,2,7, colormap[8] );
//	Vline(c, 12,3,7, colormap[2] );
//	Hline(c, 3,2,9, colormap[10] );
//	Hline(c, 3,9,9, colormap[2] );
//	Dot(c, 2, 9, colormap[3] );
//	Dot(c, 12, 2, colormap[9] );

	Vline(c, 0, 1, c->border.h-2, colormap[1]);
	Vline(c, c->border.w-1, 1, c->border.h-2, colormap[12]);
	Dot(c, 1,11, colormap[8]);
	Hline(c, 2, 11, c->window.w, colormap[9]);
	Dot(c, c->border.w-2, 11, colormap[5]);
	Vline(c, 1, 12, c->window.h, colormap[7]);
	Vline(c, c->border.w-2, 12, c->window.h, colormap[3]);
	Dot(c, 1, c->border.h-2, colormap[4]);
	Hline(c, 2, c->border.h-2, c->window.w+1, colormap[3]);
	Dot(c, 0, c->border.h-1, colormap[9]);
	Hline(c, 1, c->border.h-1, c->border.w-1, colormap[12]);
	return;


}


void dottedHline( ggi_visual_t vis, int x, int y, int h )
{
	int i;
	for( i = 0; i < h; i++ ){
		if( i % 2 ){
			ggiPutPixel(vis, x+i, y, colormap[0]);
			;//white dot
		}else{
			ggiPutPixel(vis, x+i, y, colormap[15]);
			;//black dot
		}
	} 
}

void dottedVline( ggi_visual_t vis, int x, int y, int w)
{
	int i;
	for( i = 0; i < w; i++ ){
		if( i % 2 ){
			ggiPutPixel(vis, x, y+i, colormap[0]);
			;//white dot
		}else{
			ggiPutPixel(vis, x, y+i, colormap[15]);
			;//black dot
		}
	} 
}




