/* event.c
 * event management for the display server
 */

#include "decor.h"
#include "client.h"
#include "commands.h"
#include "defines.h"
#include "menu.h"
#include "network.h"
#include "regions.h"
#include "schedule.h"
#include "visual.h"
#include "windows.h"
#include "mgwinlib.h"
#include <ggi/ggi.h>


typedef struct eventheader{
	COMMON_DATA;
}eventheader;

void store_event( client *c, ggi_event *ev )
{
	if( c == NULL )
		return;	
	if( menu( c ) ){
		menu_event( c, ev );
		return;
	}
	c->event.last++;
	if( c->event.last >= EVENTS )
		c->event.last = 0;
	c->event.buffer[c->event.last] = *ev;   //memcpy
	printf("event of type %d stored at %d\n", ev->any.type, c->event.last);
	if ( ev->any.type == 5 )
	  {
	    printf( "gii_key_event.modifiers = %d\n", ev->key.modifiers );
	    printf( "gii_key_event.sym = %d\n", ev->key.sym );
	    printf( "gii_key_event.label = %d\n", ev->key.label );
	    printf( "gii_key_event.button = %d\n", ev->key.button );
	  }
}

void get_event( client *c, gii_event *ev )
{
	int i = c->event.last + 1;
	while(1){
		if( i >= EVENTS )
			i = 0;
		if( (1 << c->event.buffer[i].any.type) & c->event.mask){

			memcpy(ev, &c->event.buffer[i],
			       c->event.buffer[i].any.size);

			c->event.buffer[i].any.type = 0;
	//printf("event of type %d released from %d \n", ev->any.type, i );
			return;
		}
		if( i == c->event.last ){
			ev->any.type = Nothing;
			ev->any.size = 0;
			return;
		}
		i++;
	}
}


void put_uint8(unsigned char *data, int offset, uint8 i)
{
	*(data+offset) = i & 0xff;
}

void put_uint32(unsigned char *data, int offset, uint32 i)
{
	*(data + offset) = (i >> 24) & 0xff;
	*(data + offset+1) = (i >> 16) & 0xff;
	*(data + offset+2) = (i >> 8) & 0xff;
	*(data + offset+3) = (i) & 0xff;
//	printf(" i = %d -> %d %d %d %d\n", i, *(data+offset),
//	       *(data+offset+1), *(data+offset+2), *(data+offset+3));
}

void put_sint32(unsigned char *data, int offset, sint32 i)
{
	int j = abs(i);
	*(data + offset) = (j >> 24) & 0xff;
	*(data + offset+1) = (j >> 16) & 0xff;
	*(data + offset+2) = (j >> 8) & 0xff;
	*(data + offset+3) = (j) & 0xff;
	if( j < 0 )
		*(data + offset) |= 0x80;
}
uint8 get_uint8(unsigned char *data, int offset)
{
        return *(data +offset);
}

uint32 get_uint32(unsigned char *data, int offset)
{
        unsigned int x1, x2, x3, x4;
        x1 = *(data + offset) << 24;
        x2 = *(data + offset + 1) << 16;
        x3 = *(data + offset + 2) << 8;
        x4 = *(data + offset + 3);
        return x1 + x2 + x3 + x4;
}

sint32 get_sint32(unsigned char *data, int offset)
{
        int i, x1, x2, x3, x4;
        x1 = (*(data + offset) & 0x7f) << 24;
        x2 = *(data + offset + 1) << 16;
        x3 = *(data + offset + 2) << 8;
        x4 = *(data + offset + 3);
        i = x1+x2+x3+x4;
        if( *(data + offset) & 0x80 )
                i = -i;
        return i;
}


int get_winmgevent(ggi_event *event, winmg_event *ev)
{
	unsigned char *data;
	if( event->any.type != evInformation )
		return -1;
	data = event->cmd.data;
        ev->any.size = get_uint8(data, 0);
        ev->any.type = get_uint8(data, 1);
        switch(ev->any.type){
        case evMouse:
                ev->mouse.direction = get_uint8(data, 2);
                break;
        case evFocus:
                ev->focus.lost = get_uint32(data, 2);
                ev->focus.new = get_uint32(data, 6);
                break;
        case evMove:
                ev->move.x = get_sint32(data, 2);
                ev->move.y = get_sint32(data, 6);
                break;
        case evResize:
                ev->resize.w = get_uint32(data, 2);
                ev->resize.h = get_uint32(data, 6);
                break;
        }
        return 0;
}

void store_winmg_event(client *c, winmg_event *ev)
{
	unsigned char *data;
	gii_event target;
	// event header + ev.cmd.code
	target.any.size =  sizeof(eventheader) + 4 +  ev->any.size;
	target.any.type = evInformation;
	target.any.error = 0;
	target.any.origin = 0;
	target.any.target = 0;
	gettimeofday(&target.any.time, NULL);
	target.cmd.code = 0xffffffff;
	data = target.cmd.data;
	put_uint8(data, 0, ev->any.size);
	put_uint8(data, 1, ev->any.type);
	switch(ev->any.type){
	case evMouse:
//		printf("evMouse\n");
		put_uint8(data, 2, ev->mouse.direction);
		break;
	case evFocus:
//		printf("evFocus: %d -> %d\n", ev->focus.lost, ev->focus.new);
		put_uint32(data, 2, ev->focus.lost); 
		put_uint32(data, 6, ev->focus.new); 
		break;
	case evMove:
//		printf("evMove\n");
		put_sint32(data, 2, ev->move.x); 
		put_sint32(data, 6, ev->move.y);
		break;
	case evResize:
//		printf("evResize\n");
		put_uint32(data, 2, ev->resize.w); 
		put_uint32(data, 6, ev->resize.h); 
		break;
	default: 
		printf("error\n");
		break;
	}
	store_event(c, &target);

}

void send_mouse_event(client *c, int direction)
{
	winmg_event ev;
	ev.any.size = 3;
	ev.any.type = evMouse;
	ev.mouse.direction = direction;
	store_winmg_event(c, &ev);
}

void send_move_event(client *c, int x, int y)
{
	winmg_event ev;
	ev.any.size = 10;
	ev.any.type = evMove;
	ev.move.x = x;
	ev.move.y = y;
	store_winmg_event(c, &ev);
}

void send_resize_event(client *c)
{
	winmg_event ev;
	ev.any.size = 10;
	ev. any.type = evResize;
	ev.resize.w = c->new.w - (border(c, left) + border(c, right));
	ev.resize.h = c->new.h - (border(c, top) + border(c, bottom ));
	store_winmg_event(c, &ev);
}

void broadcast_focus_event(int new, int lost)
{
	client *c;
	winmg_event ev;
	ev.any.size = 10;
	ev.any.type = evFocus;
	ev.focus.new = new;
	ev.focus.lost = lost;
	c = client_list();
	while( c != NULL ){
		store_winmg_event(c, &ev);
		c = c->next;
	}
}	


//{todo speed up scan by not searching if there has been no change }
ggi_event_mask scan_events(client *c, ggi_event_mask mask )
{
	ggi_event_mask ev;
	int i;
	ev = 0;
	for( i = 0; i < EVENTS; i++ ){
		ev |= 1 << c->event.buffer[i].any.type;
	}
	ev &= mask;
//	printf("available events match %x\n", ev);
	return ev;
}

// set global event mask
void set_eventmask( )
{
	ggi_event_mask ev = 0;
	client *c = client_list();
	while( c != NULL ){
		ev |= c->event.mask;
		c = c->next;
	}
	ggiSetEventMask( target_visual(), ev );
}


void broadcast_event( ggi_event *event )
{
	client *c = client_list();
	while( c != NULL ){
		store_event( c, event);		
		c = c->next;
	}
}


// simple eventpoll if needed
// {todo - read all events!}
void handle_events(struct timeval *timeout)
{
	ggi_event_mask evmask;
	ggi_event event;
	evmask = ggiEventPoll(target_visual(), emAll, timeout );
	if( evmask == 0)
		return;
	ggiEventRead( target_visual(), &event, evmask );
	switch( event.any.type ){
	case evNothing: 
		printf("evNothing\n");
		break;		
	case evCommand:
		printf("evCommand\n");
		break;
	case evInformation: 
		broadcast_event( &event );
//		printf("evInformation\n");
		break;
	case evExpose:
	{
		region r;
		r.x = event.expose.x;   r.y = event.expose.y;
		r.w = event.expose.w;   r.h = event.expose.h;
		printf("redraw from handle_events\n");
		redraw_screen_area( &r );
		break;
	}
	case evKeyPress:
	case evKeyRelease:
	case evKeyRepeat:
		store_event( get_focus(), &event);
		break;
	case evPtrRelative:
	{
		// rebuild to absolute pointer movement.
		pointer_position old;
		get_pointer( &old );
		event.any.type = evPtrAbsolute;
		event.pmove.x += old.x;
		event.pmove.y += old.y;
		event.pmove.z += old.z;
		event.pmove.wheel += old.wheel;
	}
	case evPtrAbsolute:
	{
		pointer_move(event.pmove.x, event.pmove.y, 
			     event.pmove.z, event.pmove.wheel);
		if( on_evtarget()){
			event.pmove.x -= curwin()->window.x;
			event.pmove.y -= curwin()->window.y;
			store_event( curwin(),  &event );
		}
		break;
	}
	case evPtrButtonPress:
	{		
		if( on_evtarget())
			store_event( curwin(), &event );
		else 
			PtrButtonPressed(event.pbutton.button);
		break;
	}
	case evPtrButtonRelease:
	{
		if( on_evtarget() )
			store_event( curwin(), &event );
		else
			PtrButtonReleased(event.pbutton.button);
		break;
	}
	case evValRelative:
		printf("evValRelative\n");
		break;
	case evValAbsolute:
		printf("evValAbsolute\n");
		break;
	default:
		printf("other event\n");
		break;
	}
}


//{todo - check eventfd if there is one}
void process_events( fd_set *read_set)
{
	struct timeval timeout = {0,0};
	handle_events(&timeout);
}

// event thread if threading is used!
void *event_main( void* arg )
{
	struct timeval timeout;
	// not quit not yet defined - should be a flag or something!
	while( not_quit() ){
		timeout.tv_usec = 0;
		timeout.tv_sec = AFTERLIFE;
		handle_events( &timeout );
	}
	return NULL;
}
















