/* libggi.c
 * MG server
 * routines for interfacing to libggi
 */

#include "commands.h"
#include "network.h"
#include "client.h"
#include "windows.h"
#include "memory.h"
#include "regions.h"
#include <stdlib.h>




void load_gc( client *c )
{
	ggiSetGCForeground( c->visual.target, c->visual.fg_color );
	ggiSetGCBackground( c->visual.target, c->visual.bg_color );
	// clip rectangle can be ignored since it is changed with
	// every drawing operation
}


void setMGClipping( client *c, region *clip )
{
	region r;
	int x0, y0, x1, y1;
	r.x = c->visual.cliptl.x + c->window.x;
	r.y = c->visual.cliptl.y + c->window.y;
	r.w = c->visual.clipbr.x - c->visual.cliptl.x;
	r.h = c->visual.clipbr.y - c->visual.cliptl.y;
	x0 = MAX( r.x, clip->x );
	y0 = MAX( r.y, clip->y );
	x1 = MAX( x0, MIN( r.x + r.w, clip->x + clip->w ) );
	y1 = MAX( y0, MIN( r.y + r.h, clip->y + clip->h ) );
//	printf("clipping: %d %d %d %d\n", x0, y0, x1, y1 );
	ggiSetGCClipping( c->visual.target, x0, y0, x1, y1 );	
}

/* make sure client and server are in sync */
void mg_flush( client *c )
{
	MG_packet packet;
	ggiFlush( c->visual.target );
	init_packet( &packet, MGflush );
	queue_packet(&c->comm, &packet );	
}


void mg_setflags( client *c, ggi_flags *f )
{
	c->visual.flags = *f;
	// {todo} update window with new flags
	// set async/ sync mode
}

void mg_getmode( client *c )
{
	MG_packet packet;
	init_packet( &packet, MGgetmode );
	store_ggimode( &packet, &c->visual.mode );
	queue_packet(&c->comm, &packet);
}


int checkmode( client *c, ggi_mode *tm )
{
	int err = 0;
	ggi_mode target_mode;
	ggiGetMode(c->visual.target, &target_mode);

	// check graphtype - ok or emulated?
	// simple code for now: allow only the same as the visual.target has
	if( tm->graphtype == GGI_AUTO )
		tm->graphtype = target_mode.graphtype;
	
	if( tm->graphtype != target_mode.graphtype )
		err++;
	
	// calculate size:
	// virtual is ggi_auto, but not visible
	if( tm->virt.x == GGI_AUTO && tm->visible.x != GGI_AUTO ){
		tm->virt.x = tm->visible.x;
	}

	if( tm->virt.y == GGI_AUTO && tm->visible.y != GGI_AUTO ){
		tm->virt.y = tm->visible.y;
	}
	
	// visible is ggi_auto but not virtual
	if( tm->virt.x != GGI_AUTO && tm->visible.x == GGI_AUTO ){
		tm->visible.x = tm->virt.x;
	}

	if( tm->virt.y != GGI_AUTO && tm->visible.y == GGI_AUTO ){
		tm->visible.y = tm->virt.y;
	}

	// both ar ggi_auto
	if( tm->virt.x == GGI_AUTO && tm->visible.x == GGI_AUTO ){
		if( c == desktop() )
			tm->visible.x = tm->virt.x = screen_width();
		else
			tm->visible.x = tm->virt.x = env_width(c);

	}

	if( tm->virt.y == GGI_AUTO && tm->visible.y == GGI_AUTO ){
		if( c == desktop() )
			tm->visible.y = tm->virt.y = screen_height();
		else
			tm->visible.y = tm->virt.y = env_height(c);
	}
	
	// too big?
	// truncate #frames
	while( (needed_memory( tm ) > available_memory()) && tm->frames > 0){
		tm->frames--;
		err++;
		printf("truncating frames\n");
	}
	
	// truncate virtual -> visible
	if( needed_memory(tm) > available_memory() ){
		tm->virt.x = tm->visible.x;
		tm->virt.y = tm->visible.y;
		err++;
		printf("truncating virtual\n");
	}

	// still too big?
	// truncate both, while keeping proportions
	while( needed_memory(tm) > available_memory() ){
		tm->virt.x = tm->visible.x = tm->visible.x / 2;
		tm->virt.y = tm->visible.y = tm->visible.y / 2;
		err++;
		printf("truncating visible\n");
	}
	return err;
}

void mg_checkmode( client *c, ggi_mode *tm )
{
	MG_packet packet;
	int err;
	printf("mg_checkmode\n");
	err = checkmode(c, tm);
	init_packet( &packet, MGcheckmode );
	store_sint32( &packet, err );
	store_ggimode( &packet, tm );
	queue_packet( &c->comm, &packet );
}

void mg_setmode( client *c, ggi_mode *tm )
{
	MG_packet packet;
	int err;
	if( !(err = checkmode( c, tm )) ){
		reshape_window(c, tm);
	}
	init_packet( &packet, MGsetmode );
	store_sint32( &packet, err );
	store_ggimode( &packet, tm );
	queue_packet(&c->comm, &packet );
}


// {todo} use different palettes for each window
void mg_mapcolor( client *c, ggi_color *col )
{
	MG_packet packet;
	ggi_pixel pix = ggiMapColor(c->visual.target, col);
	init_packet(&packet, MGmapcolor );
	store_uint32(&packet, pix );
	queue_packet( &c->comm, &packet );
}

void mg_unmappixel( client *c, unsigned int pix )
{
	MG_packet packet;
	ggi_color col;
	ggiUnmapPixel( c->visual.target, pix, &col );
	init_packet( &packet, MGmapcolor );
	store_ggicolor( &packet, &col );
}

void mg_setpalvec(client *c, int s, int len, ggi_color *cmap)
{
	//{todo} ge varje window separata paletter
	ggiSetPalette( c->visual.target, s, len, cmap );
}


void mg_getpalvec(client *c, int s, int len)
{
	MG_packet packet;
	int i;
	ggi_color *cmap;
	cmap = malloc( sizeof(ggi_color)*len );
	ggiGetPalette( c->visual.target, s, len, cmap );
	init_packet( &packet, MGgetpalvec );
	for( i = 0; i < len; i++ )
		store_ggicolor( &packet, cmap + i );
	queue_packet( &c->comm, &packet );
	free( cmap );
}

void mg_setorigin( client *c, int x, int y )
{
	MG_packet packet;
	int err = 0;
	if( x < 0 || x > c->visual.mode.virt.x ||
	    y < 0 || y > c->visual.mode.virt.y )
		err = 1;
	else{
		c->visual.origin.x = x;
		c->visual.origin.y = y;
		redraw_window( c );
	}
	init_packet( &packet, MGsetorigin );
	store_sint32( &packet, err );
	queue_packet( &c->comm, &packet );
}


void mg_fillscreen( client *c )
{
	region *r;
	r = c->window_clip;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiFillscreen(c->visual.target);
		r = r->next;
	}
	unlock_windows();
	// {todo} draw on buffer
}

void mg_putc( client *c, int x, int y, char ch)
{
	region *r;
	x = x+c->window.x;
	y = y+c->window.y;
	r = c->window_clip;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiPutc(c->visual.target, x, y, ch);
		r = r->next;
	}
	unlock_windows();
  // {todo} draw on buffer 
}

void mg_puts( client *c, int x, int y, char *str )
{
	region *r;
	x = x+c->window.x;
	y = y+c->window.y;
//        printf("x = %d, y = %d, str = %s\n", x, y, str);
	load_gc( c );
	r = c->window_clip;
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiPuts(c->visual.target, x, y, str);
		r = r->next;
	}
	unlock_windows();
	//{todo} draw on buffer
}

void mg_drawpixel( client *c, int x, int y )
{ 
	region *r;
	x = x+c->window.x;
	y = y+c->window.y;
	r = c->window_clip;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		if( in_region( x, y, r )){
			setMGClipping(c, r );
			ggiDrawPixel(c->visual.target, x, y);
			break;
		}
		r = r->next;
	}
	unlock_windows();
	// {todo}
}

void mg_putpixel( client *c, int x, int y, ggi_pixel pixel)
{
	region *r;
	x = x+c->window.x;
	y = y+c->window.y;
	r = c->window_clip;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		if( in_region( x, y, r )){
			setMGClipping(c, r );
			ggiPutPixel(c->visual.target, x, y, pixel);
			break;
		}
		r = r->next;
	}
	unlock_windows();
	// {todo}
}

/* gammal funktion
int in_region(int x, int y, region *r)
{
	return x > r->x && x < r->x + r->w && y > r->y && y < r->y + r->h;
}
*/

void mg_getpixel( client *c, int x, int y)
{
	MG_packet packet;
	region *r  = c->window_clip;
	x = x+c->window.x;
	y = y+c->window.y;
	lock_windows();
	while( r != NULL ){
		if( in_region( x, y, r ) ){
			ggi_pixel pix;
			ggiGetPixel(c->visual.target, x, y, &pix);
			init_packet( &packet, MGgetpixel );
			store_uint16(&packet, 0);
			store_uint32(&packet, pix );
			queue_packet(&c->comm, &packet);
			return;
		}
		r = r->next;
	}
	unlock_windows();
	init_packet(&packet, MGgetpixel );
	store_uint16(&packet, 1);
	store_uint32(&packet, 0 );
	queue_packet(&c->comm, &packet);
	// {todo} handle buffers
}

void mg_drawline( client *c, int x, int y, int xe, int ye )
{
	region *r;
	x = x+c->window.x;  xe = xe+c->window.x;
	y = y+c->window.y;  ye = ye+c->window.y;
	load_gc( c );
	r = c->window_clip;
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiDrawLine(c->visual.target, x, y, xe, ye);
		r = r->next;
	}
	unlock_windows();
	// todo
}

void mg_drawhline( client *c, int x, int y, int h )
{
	region *r;
	x = x+c->window.x;
	y = y+c->window.y;
	load_gc( c );
	r = c->window_clip;
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiDrawHLine(c->visual.target, x, y, h);
		r = r->next;
	}
	unlock_windows();
	// todo
}

void mg_puthline( client *c, int x, int y, int w, void *buffer )
{
	region *r;
	x = x + c->window.x;
	y = y + c->window.y;
	load_gc( c );
	r = c->window_clip;
	lock_windows();
	while( r != NULL ){
		setMGClipping( c, r);
		ggiPutHLine( c->visual.target, x, y, w, buffer );
		r = r->next;
	}
	unlock_windows();
}

void mg_putvline( client *c, int x, int y, int h, void *buffer )
{
	region *r;
	x = x + c->window.x;
	y = y + c->window.y;
	load_gc( c );
	r = c->window_clip;
	lock_windows();
	while( r != NULL ){
		setMGClipping( c, r);
		ggiPutVLine( c->visual.target, x, y, h, buffer );
		r = r->next;
	}
	unlock_windows();
}



void mg_drawvline( client *c, int x, int y, int h )
{
	region *r = c->window_clip;
	x = x+c->window.x;
	y = y+c->window.y;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiDrawVLine(c->visual.target, x, y, h);
		r = r->next;
	}
	unlock_windows();
	// todo
}


void mg_copybox( client *c, int x, int y, int w, int h, int nx, int ny)
{
	region *r, *cut1, *cut2;
	region box, clip;
	int dx, dy;
	box.x = x + c->window.x;      
	box.y = y + c->window.y;
	box.w = w;
	box.h = h;
	box.next = clip.next = NULL;
	dx = nx - x;
	dy = ny - y;
       
	cut1 = intersect( &box, c->window_clip );
	translate( cut1, dx, dy );
	cut2 = intersect( cut1, c->window_clip );
	translate( cut2, -dx, -dy );
	r = cut2;
	while( r != NULL ){
		int err;
		clip.x = MIN( r->x, r->x + dx );
		clip.y = MIN( r->y, r->y + dy );
		clip.w = r->w + abs(dx);
		clip.h = r->h + abs(dy);
		setMGClipping(c, &clip);
		err = ggiCopyBox(c->visual.target, r->x, r->y, r->w, r->h, 
				 r->x + dx, r->y + dy );
		r = r->next;
	}
	destroy_regions( &cut1 );
	destroy_regions( &cut2 );
}


void mg_drawbox( client *c, int x, int y, int w, int h )
{  
	region *r = c->window_clip;
	x = x+c->window.x;
	y = y+c->window.y;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiDrawBox(c->visual.target, x, y, w, h);
		r = r->next;
	}
	unlock_windows();
	// todo
}

void mg_putbox( client *c, int x, int y, int w, int h, void *buffer )
{
	region *r = c->window_clip;
	x = x+c->window.x;
	y = y+c->window.y;
	load_gc( c );
	lock_windows();
	while( r != NULL ){
		setMGClipping(c, r );
		ggiPutBox(c->visual.target, x, y, w, h, buffer);
		r = r->next;
	}
	unlock_windows();
	// todo
}


void mg_getbox( client *c, int x, int y, int w, int h, int l)
{
	MG_packet packet;
	void *buffer = malloc( l );
	x += c->window.x;
	y += c->window.y;
	ggiGetBox(c->visual.target, x, y, w, h, buffer );
	init_packet( &packet, MGgetbox );
	store_bytes( &packet, buffer, l );
	queue_packet(&c->comm, &packet);
	free( buffer );
}

void mg_gethline( client *c, int x, int y, int w, int l )
{
	MG_packet packet;
	void *buffer = malloc( l );
	x += c->window.x;
	y += c->window.y;
	ggiGetHLine(c->visual.target, x, y, w,  buffer );
	init_packet( &packet, MGgethline );
	store_bytes( &packet, buffer, l );
	queue_packet(&c->comm, &packet);
	free( buffer );
}

void mg_getvline( client* c, int x, int y, int w, int l )
{
	MG_packet packet;
	void *buffer = malloc( l );
	x += c->window.x;
	y += c->window.y;
	ggiGetVLine(c->visual.target, x, y, w,  buffer );
	init_packet( &packet, MGgetvline );
	store_bytes( &packet, buffer, l );
	queue_packet(&c->comm, &packet);
	free( buffer );
}
