#include "client.h"
#include "event.h"
#include "decor.h"
#include "libggi.h"
#include "windows.h"
#include <stdlib.h>
#include <string.h>

typedef struct{
	char label[32];
	int x, y, w, h;
	client *menu;
	int state;  /* 0 - untouched, 1- hovering 2- pushed */
	void (*action)(client *);
} button;



#define BUTTON_WIDGET 1

typedef struct _widget{
	struct _widget *next;
	int type;
	button *b;
} widget;


struct menu_info{
	widget *widgets;
	widget *current;
};


/* create a window menu */
client *create_menu( client *c )
{
	client *menu = malloc( sizeof( client ));
	init_fields( menu, 0 );
	strncpy( menu->name, "Window Operations", 32);
	menu->private = malloc( sizeof( struct menu_info));
	((struct menu_info *)(menu->private))->current = NULL;
	((struct menu_info *)(menu->private))->widgets = NULL;
	menu->parent = c;
	menu->border.x = menu->window.x = c->window.x;
	menu->border.y = menu->window.y = c->window.y;
	menu->flags |= NODECOR | FLOAT;;
	insert( menu, FIRST );
	return menu;
}

extern void deactivate_menu( void );

void destroy_menu(client *m )
{
	widget *w1, *w2;
	w1 = w2= ((struct menu_info*)m->private)->widgets;
	while( w1 != NULL ){
		w1 = w1->next;
		free( w2 );
		w2 = w1;
	}
	free(m->private);
	remove_client( m );  // borde ge felutskrift (socket closed)
	deactivate_menu();
}


// return true if c is a menu;
int menu( client *c )
{
	if( c == NULL )
		return 0;
	return (int)c->parent;
}

// draw the button b;
void draw_button( button *b )
{ 
	int fg, bg;
	fg = 15; bg = 2;
	if( b->state)
		bg = 5;
 
	set_gc( b->menu, bg, bg );
	mg_drawbox( b->menu, b->x, b->y, b->w, b->h );

	set_gc( b->menu, fg, bg );
	mg_puts( b->menu, b->x+2, b->y+2, b->label );
	if( b->state > 0 ){
		set_gc( b->menu, bg-2, bg-2 );
		mg_drawvline( b->menu, b->x, b->y, b->h );
		mg_drawhline( b->menu, b->x, b->y, b->w );
		mg_drawvline( b->menu, b->x+b->w, b->y, b->h );
		mg_drawhline( b->menu, b->x, b->y+b->h, b->w );
	}
	ggiFlush( b->menu->visual.target );
}

void enter_button( button *b )
{
        b->state = 1;
	printf("enter %s\n", b->label );
	draw_button( b );
}

void exit_button( button *b )
{
	b->state = 0;
	printf("exit %s\n", b->label );
	draw_button( b );
}

void push_button( button *b )
{
	b->state = 2;
	draw_button( b );
}

void release_button( button *b )
{
	b->state = 1;
	draw_button( b );
}



/* fiffa support fr border i add_button */
void add_button( client *menu, button *b)
{
	int y;
	widget *w, *n;
	/* resize window */
	change_window( menu, menu->border.x, menu->border.y, 
		       MAX( b->w, menu->border.w ),
		       menu->border.h + b->h );	
	menu->visual.cliptl.x = 0;
	menu->visual.cliptl.y = 0;
	menu->visual.clipbr.x = MAX( b->w, menu->border.w );
	menu->visual.clipbr.y = menu->border.h + b->h;
	/* change x & y to fit position */
	w = ((struct menu_info *)menu->private)->widgets;
	if( w == NULL )
		y = 0;
	else
		y = w->b->y + w->b->h;
	b->y = y;
	b->menu = menu;
	/* add button to list */
	n = (widget *)malloc( sizeof( widget ) );
	n->next = w;
	n->type = BUTTON_WIDGET;
	n->b = b;
	((struct menu_info *)menu->private)->widgets = n;
}

button *create_button( const char *label, void (*action)(client *) )
{
	button *b = (button *)malloc( sizeof( button ) );
	b->x = 0;
	b->y = 0;
	b->w = 60;
	b->h = 12;
	b->state = 0;
	b->action = action;
	strncpy( b->label, label, 32 );
	return b;
}


void extend_menu( client *menu, const char *label, void (*action)(client *) )
{
	button *b = create_button( label, action );
	add_button( menu, b );
//	draw_button( b );
}



int on_button( button *b, int x, int y )
{
	return (x > b->x &&
		x <= b->x + b->w &&
		y > b->y &&
		y <= b->y + b->h );
}

void menu_event( client *menu, gii_event *ev )
{
	switch( ev->any.type ){
	case evNothing:
	case evCommand: break;
	case evInformation:{
		winmg_event mgev;
		if( get_winmgevent( ev, &mgev ) )
			break;
		if( mgev.any.type == evMouse )	
			if( !mgev.mouse.direction ){
				button *b;
				b = ((struct menu_info*)menu->private)->current->b;
				b->state = 0;
				printf("exit from %s\n", b->label);
				((struct menu_info*)menu->private)->current = NULL;
				draw_button( b );
			}
		break;
	}		
	case evExpose:{
		//redraw all buttons;
		widget *w;
		w = ((struct menu_info*)menu->private)->widgets;
		while( w != NULL ){
			draw_button( w->b );
			w = w->next;
		}
		break;
	}
	case evKeyRelease:
		// call action;
		break;
      	case evPtrAbsolute:{
		// kolla om annan button;
		widget *w, *current;
		w = ((struct menu_info*)menu->private)->widgets;
		current = ((struct menu_info*)menu->private)->current;
		while( w != NULL ){
			if( on_button( w->b, ev->pmove.x, ev->pmove.y ) ){
				// perform enter/ exit
				if( w != current ){
					if( current != NULL )
						exit_button( current->b );
					if( w != NULL )
						enter_button( w->b );
					((struct menu_info*)menu->private)->current=w;
				}
				break; //while
			}
			w = w->next;
		}
		break;
	}
	case evPtrButtonRelease:{
		void (*action)(client *);
		client *c =  menu->parent;
		widget *current = ((struct menu_info*)menu->private)->current;
		action = current->b->action;
		
		if( current != NULL ){
			// destroy menu:
			destroy_menu( current->b->menu );
			update_focus();
			action( c );
		}
		break;
	}
	default: break;		
	}
}





