#ifndef MGWINLIB_H
#define MGWINLIB_H
#include <ggi/ggi.h>


/* the window manager events are encoded in gii Information events */

enum winmg_event_type{
	evBad = 0,
	evMouse,
	evFocus,
	evMove,
        evResize
}winmg_event_type;


#define COMMON_WINMG_DATA \
        uint8 size;          /* size of event in bytes */\
        uint8 type          /* type of this event     */

typedef struct winmg_any_event{
	COMMON_WINMG_DATA;
}winmg_any_event;

typedef struct winmg_resize_event{
	COMMON_WINMG_DATA;
	uint32 w, h;       /* requested new size */
}winmg_resize_event;

typedef struct winmg_move_event{
	COMMON_WINMG_DATA;
	sint32 x, y;       /* new position */
}winmg_move_event;

typedef struct winmg_focus_event{
	COMMON_WINMG_DATA;
	uint32 lost;     /* previous focus */
	uint32 new;      /* new focus */
}winmg_focus_event;

typedef struct winmg_mouse_event{
	COMMON_WINMG_DATA;
	uint8 direction;       // 0 if mouse leaves, 1 if mouse enters visual
}winmg_mouse_event;

typedef union winmg_event{
	uint8                 size;
	winmg_any_event       any;
	winmg_resize_event    resize;
	winmg_move_event      move;
	winmg_focus_event     focus;
	winmg_mouse_event     mouse;
}winmg_event;


enum winmg_commands{
	WINMGgetpos,
	WINMGsetpos,
	WINMGraise,
	WINMGlower,
        WINMGsetflags,
	WINMGgetflags
}winmg_commands;


typedef struct winmg_cmd{
	uint8 command;
	sint32 x0;
	sint32 x1;
	uint32 f;
}winmg_cmd;


int getwinmg_event(unsigned char *data, winmg_event *ev);
int winmg_raise( ggi_visual_t vis );
int winmg_lower( ggi_visual_t vis );
int winmg_getpos( ggi_visual_t vis, int *x, int *y );
int winmg_setpos( ggi_visual_t vis, int x, int y );
int winmg_setflags( ggi_visual_t vis, int flags );
int winmg_getflags( ggi_visual_t vis );
#endif
