/* network.h    linux-c
 * MG server
 *
 */

#ifndef NETWORK_H
#define NETWORK_H

#include <ggi/ggi.h>
//#include <ggi/internal/ggi-dl.h>
#include <pthread.h>
#include "defines.h"

#define MG_PORT 9999
#define BACKLOG 10
#define ELEMENTS 40   // 10 K of buffers!;

/* the structure MG_packet is used as a container for
 * GGI-events and commands to the system. It does not
 * allow for more than 255 different commands, and
 * 32 is already reserved for events!
 */


typedef struct MG_connection {
	int socket;
	int id;
	pthread_mutex_t wait_mutex;
	pthread_mutex_t send_mutex;
	pthread_mutex_t buffer_mutex;
	pthread_cond_t recv_status;
	int recv_flag;
	int count;
	int lastpos;
	MG_packet buffer[ELEMENTS];
} MG_connection;

void init_packet( MG_packet *packet, uint8 type );

void extract_bytes( void *dest, MG_packet *packet, int count );
void store_bytes( MG_packet *packet, const void *buf, int count );

uint8 extract_uint8( MG_packet *packet );
void store_uint8( MG_packet *packet, uint8 val );

int extract_uint16( MG_packet *packet );
void store_uint16( MG_packet *packet, int i);

int extract_sint16( MG_packet *packet );
void store_sint16( MG_packet *packet, int i );

unsigned int extract_uint32( MG_packet *packet );
void store_uint32( MG_packet *packet, unsigned int i );

int extract_sint32( MG_packet *packet );
void store_sint32( MG_packet *packet, int i );

// 4 bytes
ggi_coord extract_ggicoord( MG_packet *packet );
void store_ggicoord( MG_packet *packet, ggi_coord coord );

// 8 bytes
ggi_color extract_ggicolor( MG_packet *packet );
void store_ggicolor( MG_packet *packet, ggi_color *col );

// 24 bytes
void  extract_ggimode( ggi_mode *mode, MG_packet *packet );
void store_ggimode( MG_packet *packet, ggi_mode *mode );
  

void init_network( void );
int add_listeners( fd_set *set );

int store_giievent( MG_packet *packet, gii_event *ev );
int extract_giievent( gii_event *event, MG_packet *packet);

int send_packet( int socket, MG_packet *packet );
int recv_packet( int socket, MG_packet *packet );
int write_packet( MG_connection *conn, MG_packet *packet );

MG_packet *store_packet( MG_connection *conn, MG_packet* packet );
int wait_packet( int type, MG_connection *conn, 
		 MG_packet *packet, struct timeval *timeout );

int get_packet( int type, MG_connection *conn, MG_packet *packet );
void dump_packet( MG_packet *packet );

int queue_packet( communication *c, MG_packet *packet );
int flush_buffer( communication *c );
int load_packet( communication *c );
int accept_connection( int s );
void clear_packet( communication *c );
int new_connection( fd_set *set );

#endif







