/* request.c
 * GM server
 */

#include "mgwinlib.h"
#include "commands.h"
#include "visual.h"
#include "client.h"
#include "event.h"
#include "network.h"
#include "libggi.h"
#include "windows.h"
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>


int execute_requests(client *c)
{
	MG_packet *packet = &c->comm.packet;

	//	printf(" doing packet %d\n", packet->type );

	switch( packet->type ){

	case Nothing:
		printf("close received\n");
		return -1;
	case Command:
	{
		MG_packet reply;
		winmg_cmd command;
		command.command = extract_uint8( packet );
		command.x0 = extract_sint32( packet );
		command.x1 = extract_sint32( packet );
		command.f = extract_uint32( packet );
		execute_winmg_command( c, &command );
		init_packet( &reply, Command );
		store_uint8( &reply, command.command );
		store_sint32( &reply, command.x0 );
		store_sint32( &reply, command.x1 );
		store_uint32( &reply, command.f );
		queue_packet( &c->comm, &reply );
		break;
	}

	case MGflush:
		mg_flush(c);  // flush all pending operations
		break;
		
	case MGsetflags:
	{
		ggi_flags flags = extract_uint32( packet );
		mg_setflags(c, &flags);
		break;
	}
	
	case MGgetmode:
		mg_getmode(c);
		break;
		
	case MGcheckmode:
	{
		ggi_mode mode;
		extract_ggimode( &mode, packet );
		mg_checkmode(c, &mode);
		break;
	}
	
	case MGsetmode:
	{
		ggi_mode mode;
		extract_ggimode( &mode, packet );
		mg_setmode(c, &mode);
		break;
	}
	
	case MGeventread:
	{
		MG_packet reply;
		gii_event ev;
//		printf("event_read received...\n");
		init_packet( &reply, MGeventread );
		get_event( c, &ev );

		store_giievent( &reply, &ev);
/*		if(ev.any.type != 0){
			printf(" event size = %d\n", ev.any.size );
//			printf(" ...sending event of type %d\n", ev.any.type);
//			dump_packet( &reply );
			
		}
*/
		queue_packet( &c->comm, &reply );
		break;
	}

	case MGseteventmask:
		c->event.mask = (ggi_event_mask)extract_uint32( packet );
		set_eventmask();
		break;
	
	case MGmapcolor:
	{
		ggi_color col = extract_ggicolor( packet );
		mg_mapcolor(c, &col);
		break;
	}
	
	case MGunmappixel:
	{
		unsigned int pix=extract_uint32(  packet );
		mg_unmappixel(c, pix);
		break;
	}
	
	case MGsetpalvec:
	{
		int i, s, len;
		ggi_color *cmap;
		s = extract_sint32( packet );
		len = extract_uint8( packet );
		cmap = (ggi_color *)malloc( sizeof(ggi_color)*len );
		for( i = 0; i < len; i++)
			*(cmap + i ) = extract_ggicolor( packet );
		mg_setpalvec( c, s, len, cmap);
		free( cmap );
		break;
	}
        
	case MGgetpalvec:
	{
		int s, len;
		s = extract_sint32( packet );
		len = extract_uint8( packet );
		mg_getpalvec( c, s, len );
	}

	case MGsetorigin:
	{
		int x, y;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		mg_setorigin(c, x, y );
	}
    
	case MGfillscreen:
		mg_fillscreen(c);
		break;

	case MGputc:
	{
		int x,y;
		char ch;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		ch = extract_uint8( packet );
		mg_putc(c, x, y, ch);
		break;
	}

	case MGputs:
	{
		int x, y, len;
		char *str;
		len = extract_uint8(  packet );
		x = extract_sint32( packet );
		y = extract_sint32(  packet );
		str = (char *)malloc( len + 1);
		extract_bytes( str, packet, len );
		*(str+len) = '\0';
		mg_puts( c, x, y, str );
		free( str );
		break;
	}
	
	
	case MGdrawpixel:
	{
		int x, y;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		mg_drawpixel( c, x, y );
		break;
	}
	
	case MGputpixel:
	{
		int x,y,col;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		col = extract_uint32( packet );
		mg_putpixel(c, x, y, col );
		break;
	}
	
	case MGgetpixel:
	{
		int x,y;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		mg_getpixel(c, x, y);
		break;
	}
	
	case MGdrawline:
	{
		int x,y,xe,ye;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		xe = extract_sint32( packet );
		ye = extract_sint32( packet );
		mg_drawline(c, x, y, xe, ye);
		break;
	}

	case MGdrawhline:
	{
		int x,y,h;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		h = extract_sint32( packet );
		mg_drawhline(c, x, y, h );
		break;
	}

	case MGputhline:
	{
		int x, y, w, len;
		void *buffer;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_uint8( packet );
		len = extract_uint8( packet );  
//		printf("puthline %d %d %d %d\n", x,y,w,len);
		buffer = (void *)malloc( len );
		extract_bytes( buffer, packet, len );
		mg_puthline(c, x, y, w, buffer);
		free( buffer );
		break;
	}
	
	case MGgethline:
	{
		int x, y, w, len;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_uint8( packet );
		len = extract_uint8( packet );
		mg_gethline(c, x, y, w, len );
		break;
	}

	case MGdrawvline:
	{
		int x,y,h;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		h = extract_sint32( packet );
		mg_drawvline(c, x, y, h );
		break;
	}
	case MGputvline:
	{
		int x, y, w, len;
		void *buffer;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_uint8( packet );
		len = extract_uint8( packet );   
		buffer = (void *)malloc( len );
		extract_bytes( buffer, packet, len );
		mg_putvline(c, x, y, w, buffer);
		free( buffer );
		break;
	}
	
	case MGgetvline:
	{
		int x, y, w, len;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_uint8( packet );
		len = extract_uint8( packet );
		mg_getvline(c, x, y, w, len );
		break;
	}

	case MGcopybox:
	{
		int x,y,w,h,nx,ny;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_sint32( packet );
		h = extract_sint32( packet );
		nx = extract_sint32( packet );
		ny = extract_sint32( packet );
		mg_copybox(c, x, y, w, h, nx, ny );
		break;
	}

	case MGdrawbox:
	{
		int x,y,w,h;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_sint32( packet );
		h = extract_sint32( packet );
		mg_drawbox(c, x, y, w, h); 
		break;
	}

	case MGputbox:
	{
		int x,y,w,h, len;
		void *buffer;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_sint32( packet );
		h = extract_sint32( packet );
		len = extract_sint32( packet );
		buffer = malloc(len);
		extract_bytes( buffer, packet, len );
		mg_putbox(c, x, y, w, h, buffer);
		free( buffer );
//		printf("putbox %d %d %d %d\n", x, y, w, h );
		break;
	}


	case MGgetbox:
	{
		int x, y, w, h, len;
		x = extract_sint32( packet );
		y = extract_sint32( packet );
		w = extract_sint32( packet );
		h = extract_sint32( packet );
		len = extract_sint32( packet );
		mg_getbox(c, x, y, w, h, len );
		break;
	}

	case MGgcchanged:
	{
		int mask = extract_uint32( packet );
		if( mask & 1 )
			c->visual.fg_color = extract_uint32( packet );
		if( mask & 2 )
			c->visual.bg_color = extract_uint32( packet );
		if(mask & 4 ){
			c->visual.cliptl = extract_ggicoord( packet );
			c->visual.clipbr = extract_ggicoord( packet );
		}
		break;
	}
	
	default:
		printf("%d not implemented\n", packet->type);
		break;
	}
	clear_packet( &c->comm );
	return 0;
}







