/* schedule.c
 * sceduler for the display server
 */

#include "event.h"
#include "client.h"
#include "defines.h"
#include "network.h"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/* global variables for the scheduler */
fd_set readable;
fd_set pending_writes;
int fdmax = 0;
int interval = 40000;
int running = 1;

/* helper functions */
void update_fdmax( int s )
{
	if( s != fdmax )
		fdmax = MAX( fdmax, s );
	else {
		int i;
		fdmax = 0;
		for( i = 0; i <= s; i++){
			if( FD_ISSET( i, &readable ) || 
			    FD_ISSET( i, &pending_writes))
				fdmax = i;
		}	
	}
}

/* functions for changing the global variables */
void add_readable( client *c )
{
	FD_SET(c->comm.socket, &readable);
	FD_SET(c->comm.socket, &pending_writes);
	update_fdmax(c->comm.socket);
}

void clear_readable( client *c )
{
	FD_CLR(c->comm.socket, &readable);
	FD_CLR(c->comm.socket, &pending_writes);
	update_fdmax(c->comm.socket);
}

void add_write( client *c )
{
	FD_SET(c->comm.socket, &readable);
	update_fdmax(c->comm.socket);
}

void clear_write( client *c)
{
	FD_CLR(c->comm.socket, &readable);
	update_fdmax(c->comm.socket);
}


/* main functions */

int not_quit( void )
{
	return running;
}

void init_scheduler( void )
{
	FD_ZERO(&readable);
	FD_ZERO(&pending_writes);
}


int frozen = 0;

void freeze( void )
{
	frozen = 1;
}

void thaw( void )
{
	frozen = 0;
}

void schedule( void )
{
	int set;
	fd_set read_set;
	fd_set write_set;
	struct timeval timeout;

	fdmax = add_listeners( &readable );
	while( not_quit() ){
		read_set = readable;
		write_set = pending_writes;
		timeout.tv_sec = 0;
		timeout.tv_usec = interval;

		set = select(fdmax + 1, &read_set, &write_set, NULL, &timeout);
		
		//{todo: add code for different event-models}

		do{
			process_events(&read_set);
		}while( frozen );
		
		if( set == 0 )
			continue;

		process_clients(&read_set, &write_set);
		
		while( (set = new_connection( &read_set )) ){
			add_client( set );
		}
	}
}




