/* windows.c
 * MG server
 *
 * the window manager
 */

#include "event.h"
#include "client.h"
#include "menu.h"
#include "windows.h"
#include "regions.h"
#include "visual.h"
#include "decor.h"
#include "schedule.h"
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <ggi/ggi.h>

int width = 600;
int height = 200;
client *cur_menu = NULL;

ggi_coord new_pos = {-1,-1};
pointer_position pointer;
ggi_pixel fill_color;
client *_resizing;
client *_moving;
client *_curwin;
client *_focus = NULL;


/* global variables to handle clicks etc */
pointer_position btndn_pos, btnup_pos;
client *btndn_win, *btnup_win;
struct timeval btndn_time;
struct timeval btnup_time;
pointer_position mof;               // move offset
struct timeval clicktime = {0,0};
int click;

void draw_outline( client *c);

void lock_windows( void )
{}

void unlock_windows( void )
{}


enum styles {OUTLINE, SOLID, SHADE};

int move_style()
{
	return SOLID;
}

void set_fillcolor( ggi_pixel col )
{
	fill_color = col;
}

int max_width( client *c)
{
  ggi_mode mode;
  ggiGetMode(target_visual(), &mode);
  return mode.virt.x - (border(c, left) + border(c, right));
  //	return width;
}

int max_height( client *c)
{
  ggi_mode mode;
  ggiGetMode(target_visual(), &mode);
 
  return mode.virt.y -(border(c,top) + border(c, bottom));

  //  return height;
}

/* should really get client env... */
int env_width(client *c)
{
  char *env = getenv("YAWS_WIDTH");
  if(env != NULL )
    return atoi(env);
  return max_width(c);
}

int env_height(client *c)
{
  char *env = getenv("YAWS_HEIGHT");
  if(env != NULL )
    return atoi(env);
  return max_height(c);
  
}


void fill_region( region *r )
{
	ggi_visual_t v = target_visual();
	ggiSetGCClipping( v, r->x, r->y, r->x + r->w, r->y + r->h );
	ggiSetGCForeground( v, fill_color );
	ggiFillscreen( v );
}

void clip_window( client *c )
{
	client *top;
	region *new;

	destroy_regions( &c->window_clip );
	new = intersect_single( &c->window, get_screen() );
	add_region( &c->window_clip, new );
	
	destroy_regions( &c->border_clip );
	new = intersect_single( &c->border, get_screen() );
	c->border_clip = difference_single( new, &c->window );
	
	top = client_list();
	while( top != c ){
		region *tmp;
		tmp=difference_single(c->window_clip, 
				      &top->border);
		destroy_regions( &c->window_clip );
		c->window_clip = tmp;
		
		tmp=difference_single(c->border_clip,
				      &top->border);
		destroy_regions( &c->border_clip );
		c->border_clip = tmp;
		top = top->next;
	}
}


// this updates all clipping lists
void update_regions( region *r )
{
	client *c = client_list();
//	printf("update_regions:\n");
	region display;
	display.x = display.y = 0;
	display.w = screen_width();
	display.h = screen_height();
	while( c != NULL ){
//		printf(" %p -> %p \n", c, c->next );
		if( covers( r, &c->border ) )
			clip_window( c );
		c = c->next;
	}
}

// send expose to client
void redraw_region( client *c, region *r)
{	
	ggi_event ev;

	region tmp = *r;
	tmp.next = NULL;
//	printf("redraw region:\n");
//	print_regions(&tmp);
	fill_region( &tmp );

	ev.expose.size = sizeof(ggi_expose_event);
	ev.expose.type = evExpose;
	ev.expose.origin = 0;
	ev.expose.target = 0;
	gettimeofday(&ev.expose.time, NULL );
	ev.expose.x = r->x - c->window.x + c->visual.origin.x;
	ev.expose.y = r->y - c->window.y + c->visual.origin.y;
	ev.expose.w = r->w;
	ev.expose.h = r->h; 
//	printf("redraw_region\n");
	store_event( c, &ev );
}



void redraw_border(client *c)
{
	region *r;
	if( c == NULL )
		return;
	r = c->border_clip;
	while( r != NULL){
		ggiSetGCClipping(c->visual.target, r->x, r->y, 
				 r->x + r->w, r->y + r->h );
		draw_decorations( c );
//		ggiDrawBox(c->visual.target, r->x, r->y, r->w, r->h );
		r = r->next;
	}
	ggiFlush(c->visual.target);
}


// paint background ( that is not covered by the desktop app )
void redraw_background( region *list )
{
	while( list != NULL ){
		fill_region( list );
		list = list->next;
	}
}


void reshape_window( client *c, ggi_mode *mode )
{
	region r;

	r.next = NULL;
	r.x = c->border.x;
	r.y = c->border.y;

	r.w = MAX( c->border.w, mode->visible.x+
		   border(c, left)+border(c, right));
	r.h = MAX( c->border.h, mode->visible.y+
		   border(c, top)+border(c, bottom));


	if( c->border.y != c->new.y )
		c->border.y = c->new.y + c->new.h - 
			(mode->visible.y +  border(c, top) + 
			 border(c, bottom ));	
	
	//printf("reshape window\n");
	//print_regions( &r );

	c->window.x = c->border.x + border(c, left);
	c->window.y = c->border.y + border(c, top);

	c->visual.mode = *mode;
	c->window.w = mode->visible.x;
	c->window.h = mode->visible.y;
	c->border.w = c->window.w + border(c, left) + border(c, right);
	c->border.h = c->window.h + border(c, top) + border(c, bottom);
	c->new = c->border;
	update_regions( &r );
//	printf("redraw from reshape_window\n");
	redraw_screen_area( &r );
}


void change_window( client *c, int x, int y, int w, int h )
{
	region r;

	r.next = NULL;
	r.x = MIN( c->window.x - border(c, left), x );
	r.y = MIN( c->window.y - border(c, top ), y );
	r.w = MAX( x+w, c->window.x+c->window.w+border(c, right) ) - r.x;
	r.h = MAX( y+h, c->window.y+c->window.h+border(c, bottom) ) - r.y;

	c->border.x = x;
	c->border.y = y;
	c->window.x = c->border.x + border(c, left);
	c->window.y = c->border.y + border(c, top);

	c->visual.mode.visible.x = w;
	c->visual.mode.visible.y = h;
	c->visual.mode.virt = c->visual.mode.visible;

	c->window.w = w;
	c->window.h = h;
	c->border.w = c->window.w + border(c, left) + border(c, right);
	c->border.h = c->window.h + border(c, top) + border(c, bottom);
	update_regions( &r );
//	printf("redraw from reshape_window\n");
	redraw_screen_area( &r );
}



void move_window(client *c, int x, int y)
{
	region r;
//	printf(">move\n");
	if( x == c->border.x && y == c->border.y )
		return;
	r.x = MIN( x, c->border.x );
	r.y = MIN( y, c->border.y );
	r.w = c->border.w + abs( x - c->border.x );
	r.h = c->border.h + abs( y - c->border.y );
	r.next = NULL;
	c->border.x = x;
	c->border.y = y;
	c->window.x = x + border(c, left);
	c->window.y = y + border(c, top);
//	printf("move window %d %d\n", x, y );
	update_regions( &r );  // update segmentation of affected windows
	if( c->border.w && c->border.h )
		redraw_screen_area( &r );   // update screen area!
//	printf("<move\n");
}

void redraw_screen_area( region *r )
{
	region *background_list = NULL;
	client *c=client_list();
	region *a;
	region screen;
	screen.x = screen.y = 0;
	screen.w = screen_width();
	screen.h = screen_height();
	if( (a = intersect_single(r, &screen )) == NULL ){
		printf("window offscreen!\n");
		return;
	}
	add_region( &background_list, create_region(a->x, a->y, a->w, a->h));
//	printf("redraw_screen_area:\n");
//	print_regions( r );
//	print_regions(background_list);
	while( c != NULL ){
		// draw content
		region *t = c->window_clip;
//		printf("clip_list: %p %p\n", c, c->next);
//		print_regions( t );
		while( t != NULL ){
			if( covers(a, t) ){
				region *tmp;
				redraw_region( c, t );
				tmp = difference_single( background_list, t );
				destroy_regions( &background_list );
				background_list = tmp;
			}
			t = t->next;
		}

		// draw border
		t = c->border_clip;
		while( t != NULL ){
			if( covers(r, t) ){
				region *tmp;
				tmp = difference_single( background_list, t );
				destroy_regions( &background_list );
				background_list = tmp;
			}
			t = t->next;
		}
		redraw_border( c );
		c = c->next;
	}
	redraw_background( background_list );
	ggiFlush(target_visual());
	destroy_regions(&a);
}

void redraw_screen( void )
{
	client *c;
	region *background_list = NULL;
	c=client_list();
	add_region( &background_list, create_region( 0,
						     0,
						     screen_width(),
						     screen_height()) );
	while( c != NULL ){
		region *tmp;
		redraw_window( c );
		tmp = difference_single( background_list, &c->border );
		destroy_regions( &background_list );
		background_list = tmp;
		c = c->next;
	}
	redraw_background( background_list );
}

void redraw_window( client *c )
{
	// draw_content
	region *t = c->window_clip;
	while( t != NULL ){
		redraw_region( c, t );
		t = t->next;
	}
	// draw border
	redraw_border( c );
}


void update_focus( void )
{
	_curwin = get_window( get_position() );
	switch( focus_policy() ){
	case FOLLOW_MOUSE:
		set_focus( _curwin );
		break;
	case LAZY:
		if( _curwin != NULL ){
			set_focus( _curwin );
		}
		break;
	default: break;
	}
}

void raise_window( client *c )
{
	if( c == NULL )
		return;
	printf(" raising %d\n",c->comm.socket);
	if( !raise_client( c ) )
		return;
	update_focus();
	update_regions( &c->border);
//	printf("redraw from raise window\n");
	redraw_screen_area( &c->border );
}

void lower_window( client *c )
{
	if( c == NULL )
		return;
	printf("lower %d\n", c->comm.socket);
	if(!lower_client( c ))
		return;
	update_focus();
	update_regions( &c->border );
//	printf("redraw from lower_window\n");
	redraw_screen_area( &c->border );
}


client *get_window( pointer_position *pos)
{
	client *c = client_list();
	while( c != NULL ){
		if( pos->x > c->border.x && 
		    pos->x < c->border.x + c->border.w &&
		    pos->y > c->border.y &&
		    pos->y < c->border.y + c->border.h ){
//			printf("window at (%d, %d) belongs to %d\n",
//			       pos->x, pos->y, c->comm.socket );
			return c;
		}
		c = c->next;
	}
	return NULL;
}


client *get_focus()
{
	return _focus;
}

int get_id(client *c)
{
	if( c == NULL )
	       return  0;
	else
		return  c->comm.socket;
}

void set_focus(client *c)
{
	client *old;

	if( menu_active() && !menu( c ) )
		return;
	if( c == _focus ){
		return;
	}

/*	if( c != NULL )
		printf("focus changed to %d\n", c->comm.socket );
	else
		printf("focus changed to NULL\n");
*/
	old = _focus;
	_focus = c;
	if( autoraise() )
		raise_window( c );
	redraw_border(old);
	redraw_border(c);
	broadcast_focus_event( get_id(c), get_id(old));
}


void display_position( void )
{
	ggi_color col;
	ggi_pixel pix;
	char buf[20];
	col.r = col.g = col.b  = 0xffff;
	sprintf(buf, "%d, %d %s", pointer.x, pointer.y, "       ");
	ggiSetGCClipping(target_visual(), 0,0, 100,10);
	pix = ggiMapColor( target_visual(), &col );
	ggiSetGCForeground( target_visual(), pix );
	ggiPuts(target_visual(), 0,0, buf);	
	ggiFlush(target_visual());
}


client *curwin()
{
	return _curwin;
}

void set_moving(client *c)
{
	if( move_style() == OUTLINE  && c != NULL )
		freeze();
	else
		thaw();
	_moving = c;
	if( c == NULL )
		update_focus();
	else
		set_focus( NULL );
	_resizing = NULL;
}

client *moving( void )
{
	return _moving;
}

void set_resizing(client *c)
{
	if( c != NULL ){
		c->new = c->border;
		ggiSetGCClipping( target_visual(), 0, 0, 
				  screen_width(), screen_height() );
		freeze();
	}
	else
	{
		thaw();
	}
	_resizing = c;
	_moving = NULL;
}

client *resizing( void )
{
	return _resizing;
}

int focus_policy( void )
{
	return LAZY;
}


int in_active( client *c, pointer_position *pos )
{
	int a;
	if( c == NULL )
		return 0;
	a = (pos->x >= c->window.x && pos->x < c->window.x + c->window.w &&
	     pos->y >= c->window.y && pos->y < c->window.y + c->window.h);
//	if( a )
//		printf("in active\n");
	return a;
}

int on_tab( client *c, pointer_position *pos )
{
	if( c == NULL )
		return 0;
	return (pos->y < c->window.y && pos->x > c->border.x+14 &&
		pos->x < c->border.x+c->border.w-14 );
}

int on_btn( client *c, pointer_position *pos )
{
	if( c == NULL )
		return 0;
	return (pos->y < c->window.y && pos->x <= c->border.x+14);
}

int on_resize( client *c, pointer_position *pos )
{
	if( c == NULL )
		return 0;
	return (pos->y < c->window.y && pos->x >= c->border.x+c->border.w-14);
}


// return click_interval in us;
int click_interval()
{
	return 500000;
}

int autoraise()
{
	return 0;
}

pointer_position *get_position( void )
{
	return &pointer;
}

void get_pointer( pointer_position *pos )
{
	memcpy(pos, &pointer, sizeof(pointer_position));
}


void set_pointer( int x, int y, int z, int wheel )
{
	pointer.x = x;  pointer.y = y;
	pointer.z = z; 	pointer.wheel = wheel;
	_curwin = get_window( &pointer );
}



ggi_pixel ggi_xor(ggi_pixel p )
{
	return p ^ 0xffffff;
}

void xor_hline( int x, int y, int w )
{
	int i;
	for( i = 0; i < w; i++ ){
		ggi_pixel p;
		ggiGetPixel( target_visual(), x+i, y, &p );
		p = ggi_xor(p);
		ggiPutPixel( target_visual(), x+i, y, p );
	}
}

void xor_vline( int x, int y, int h )
{
	int i;
	for( i = 0; i < h; i++ ){
		ggi_pixel p;
		ggiGetPixel( target_visual(), x, y+i, &p );
		p = ggi_xor(p);
		ggiPutPixel( target_visual(), x, y+i, p );
	}
}

void draw_outline( client *c)
{
	xor_hline( c->new.x, c->new.y, c->new.w );
	xor_hline( c->new.x, c->new.y+c->new.h-1, c->new.w );
	xor_vline( c->new.x, c->new.y, c->new.h );
	xor_vline( c->new.x+c->new.w-1, c->new.y, c->new.h );	
	ggiFlush( c->visual.target);
}


int visible = 0;
static void *otop = NULL;
static void *obottom = NULL;
static void *oleft = NULL; 
static void *oright = NULL;

void turn_off_outline( client *c)
{
	ggi_visual_t vis = c->visual.target;
	if( visible ){
		//putline
		ggiPutHLine( vis, c->border.x, c->border.y, c->border.w, otop);
		ggiPutHLine( vis, c->border.x, c->border.y + c->border.h,  
			     c->border.w, obottom);
		ggiPutVLine( vis, c->border.x, c->border.y, c->border.h, oleft);
		ggiPutVLine( vis, c->border.x + c->border.w, c->border.y,  
			     c->border.w, oright);
		//free space
		free( otop); free( obottom ); free( oleft ); free( oright );
		visible = 0;
	}
}

void turn_on_outline( client *c )
{
	ggi_visual_t vis = c->visual.target;
	if(!visible){
		//alloc space
		otop = malloc(c->border.w*4);
		obottom = malloc(c->border.w*4);
		oleft = malloc(c->border.h*4);
		oright = malloc(c->border.h*4);
      		//getline
		ggiGetHLine( vis, c->border.x, c->border.y, c->border.w, otop);
		ggiGetHLine( vis, c->border.x, c->border.y + c->border.h,  
			     c->border.w, obottom);
		ggiGetVLine( vis, c->border.x, c->border.y, c->border.h, oleft);
		ggiGetVLine( vis, c->border.x + c->border.w, c->border.y,  
			     c->border.w, oright);
		//drawline
		dottedHline( vis, c->border.x, c->border.y, c->border.w );
		dottedHline( vis, c->border.x, c->border.y+c->border.h, c->border.w );
		dottedVline( vis, c->border.x, c->border.y, c->border.h );
		dottedVline( vis, c->border.x+c->border.w, c->border.y, c->border.h );
		visible = 1;
	}
}


void set_outline( client *c, int x, int y, int w, int h )
{
	freeze();
	c->border.x = x;
	c->border.y = y;
	c->border.w = w;
	c->border.h = h;
	turn_on_outline( c );
}

void move_outline( client *c, int x, int y )
{
	turn_off_outline( c );
	c->border.x = x;
	c->border.y = y;
	turn_on_outline( c );
}

void resize_outline( client *c, int w, int h )
{
	turn_off_outline(c );
	c->border.w = w;
	c->border.h = h;
	turn_on_outline( c );
}


void pointer_move( int x, int y, int z, int wheel )
{
	pointer_position old;
	client *oldwin = curwin();
	get_pointer( &old );
	set_pointer( x, y, z, wheel );
	
	if( moving() != NULL ){
		if( move_style() == OUTLINE ){
			move_outline( moving(), x-mof.x, y-mof.y );
		} else {
			move_window( moving(), x-mof.x, y-mof.y );
		}
		return;
	} else if( resizing() ){
		client *r = resizing();
		draw_outline(r);		
		r->new.x = r->border.x;
		r->new.y = y > r->border.y + r->border.h - 
			border(r, top) - border(r, bottom) ? 
			r->border.y + r->border.h - 
			border(r, top) - border( r, bottom ): y ;
		r->new.w = x<r->border.x + border(r, left) + border(r, right) ?
			border(r, left) + border(r, right) : x - r->border.x ;
		r->new.h = r->border.y + r->border.h - r->new.y;
		draw_outline( r);
		return;
	}
	
	if( oldwin != curwin() ){
		send_mouse_event( oldwin, 0 );
		send_mouse_event( curwin(), 1 );
		if( (focus_policy() == FOLLOW_MOUSE) ||
		    (( focus_policy() == LAZY) && (curwin() != NULL ))){
			set_focus( curwin() );
		}
	}
}



int within_interval( struct timeval *b, struct timeval *a)
{
	int s, u;
	s = b->tv_sec - a->tv_sec;
	u = b->tv_usec - a->tv_usec;
	if( u < 0 ){
		u += 1000000;
		s--;
	}
//	printf("s = %d u = %d, ci = %d\n", s, u, click_interval());
	if( s > click_interval()/1000000 )
		return 0;
	return (s*1000000 + u) < click_interval();
}


int nearby( pointer_position pos1, pointer_position pos2 )
{
	int a, b;
	a = pos1.x - pos2.x;
	b = pos1.y - pos2.y;
	if( (a*a + b*b) < 20 )
		return 1;
	return 0;
}

int on_evtarget( void )
{
	return (in_active( curwin(), get_position()) &&
		(curwin() == get_focus() || 
		 curwin()->flags & RESPONSIVE) &&
		!(resizing()));
}



void tab_pressed( client *c, pointer_position *pos )
{
	set_moving( c );
}

void tab_released( client *c, pointer_position *pos )
{
	if( moving()){
		move_window( moving(), pos->x-mof.x, pos->y-mof.y );
		set_focus( moving() );
		set_moving( NULL );
	}
	switch( click ){
	case 0: break;
	case 1: printf("tab clicked - raise\n"); 
		raise_window( c );
		break;
	case 2: printf("tab dblclicked - lower\n"); 
		lower_window( c );
		break;
	default: printf("enough is enough! - stop clicking\n"); break;
	}
}

void perform_move( client *c )
{
	pointer_position pos;
	get_pointer( &pos );
	mof.x = pos.x - c->border.x;
	mof.y = pos.y - c->border.y;
	set_moving( c );
}

client *active_menu = NULL;

void activate_menu( client *c)
{
	active_menu = c;
	set_focus( c );
	redraw_region( c, &c->window );
}

void deactivate_menu( void )
{
	active_menu = NULL;
}

int menu_active( void )
{
	return active_menu != NULL;
}


void btn_pressed( client *c, pointer_position *pos )
{
	client *m;
	printf("btn pressed\n");
	if( menu_active() ){
		destroy_menu( active_menu );
		return;
	}
	m = create_menu( c );
	extend_menu( m, "Close", remove_client );
	extend_menu( m, "Move", perform_move );
	extend_menu( m, "Raise", raise_window );
	extend_menu( m, "Lower", lower_window );
	activate_menu( m );
}

void btn_released( client *c, pointer_position *pos )
{	
	printf("button clicked %d times\n", click);
}


void PtrButtonPressed(unsigned int button)
{
	pointer_position pos;
	get_pointer( &pos );
	if( menu_active() ){
		destroy_menu( active_menu );
		update_focus();
		return;
	}
	if( curwin() != NULL ){
		mof.x = pos.x - curwin()->border.x;
		mof.y = pos.y - curwin()->border.y;
	}

	if( on_tab( curwin(), &pos ) ){
		tab_pressed( curwin(), &pos );
	} else if( on_btn( curwin(), &pos ) ){
		btn_pressed( curwin(), &pos );
	} else if( on_resize( curwin(), &pos )){
		set_resizing( curwin() );
		draw_outline( curwin() );
	}
	
	get_pointer( &btndn_pos );
	btndn_win = curwin();
	gettimeofday( &btndn_time, NULL);
}


void PtrButtonReleased( unsigned int button )
{
	struct timeval t;
	pointer_position pos;
	get_pointer( &pos );
	gettimeofday( &t, NULL );
	
	printf("ptrbuttonreleased\n");
	if( within_interval( &t, &btndn_time ) &&
	    nearby( pos, btndn_pos)){
		if( within_interval( &t, &btnup_time ) &&
		    nearby( pos, btnup_pos )){
			click++;
		} else {
			click = 1;
		}
	} else {
		click = 0;
	}
	if( on_tab( curwin(), &pos ) ){
		tab_released( curwin(), &pos );
	} else if( on_btn( curwin(), &pos ) ){
		btn_released( curwin(), &pos );
	}
	if( moving() != NULL ){
		move_window( moving(), pos.x-mof.x, pos.y-mof.y );
		set_focus( moving() );
		set_moving( NULL );
	} else if( resizing() != NULL ){
		send_resize_event( resizing() );
		draw_outline( resizing() );
		set_resizing( NULL );
		printf("set size\n");
	}
	
	if( focus_policy() == CLICK && nearby( pos, btndn_pos ))
		set_focus( curwin() );
	memcpy( &btnup_time, &t, sizeof( struct timeval ) );
	get_pointer( &btnup_pos);
	btnup_win = curwin();
}

void adjust_window( client *c, unsigned int flags )
{
	unsigned int oldflags = c->flags;
	region r = c->border;
	r.next = 0;
	c->flags = flags;
	c->window.x = c->border.x + border(c, left);
	c->window.y = c->border.y + border(c, top);
	c->border.w = c->window.w + border(c, left) + border(c, right);
	c->border.h = c->window.h + border(c, top) + border(c, bottom);
	r.w = MAX( r.w, c->border.w );
	r.h = MAX( r.w, c->border.h );
	if( (c->flags & FLOAT) && !(oldflags & FLOAT) )
		raise_window( c);
	else
		redraw_screen_area( &r );
	ggiFlush( target_visual());
}


/* flags: 
   desktop        - window is stuck on the desktop /i.e in the bottom
   floating       - window is floting on top of other windows
   stick          - window has a fixed position and cannot be moved
   tilted         - window has label and tag on the side 
   nodecor        - window has no decorations
   icon           - window is iconified
   hidden         - window is invisible
*/ 


void execute_winmg_command(client *c,  winmg_cmd *cmd )
{
	switch(cmd->command){
	case WINMGgetpos:
		cmd->x0 = c->border.x;
		cmd->x1 = c->border.y;
		break;
	case WINMGsetpos:
		move_window( c, cmd->x0, cmd->x1 );
		break;
	case WINMGraise:
		raise_window( c );
		break;
	case WINMGlower:
		lower_window( c );
		break;
	case WINMGsetflags:
		adjust_window(c, cmd->f);
		// update everything with regard to flags
		break;
	case WINMGgetflags:
		cmd->f = c->flags;
		break;
	}
}







